#include "rdb.h"
#include <fcntl.h>
#include <unistd.h>
#include <linux/hdreg.h>
#include <sys/ioctl.h>
#include <ctype.h>
#include <stdio.h>
#include <stdlib.h>
#include <malloc.h>
#include <string.h>
#include </var/src/linux/drivers/scsi/scsi_ioctl.h>

#define BLOCKSIZE 512
#define RDBS 2000
#define PARTS 16

u_char *rdb;
int firstblock=-1;
int lastblock=0;

#define RDB(nr) ((struct RigidDiskBlock *)(rdb+(nr)*BLOCKSIZE))
#define BBB(nr) ((struct BadBlockBlock *)(rdb+(nr)*BLOCKSIZE))
#define PAB(nr) ((struct PartitionBlock *)(rdb+(nr)*BLOCKSIZE))
#define FSB(nr) ((struct FileSysHeaderBlock *)(rdb+(nr)*BLOCKSIZE))
#define LSB(nr) ((struct LoadSegBlock *)(rdb+(nr)*BLOCKSIZE))

#define BUILD3(N) (((N)[0]<<16)+((N)[1]<<8)+(N)[2])
#define BUILD2(N) (((N)[0]<<8)+(N)[1])

LONG checksum(ptr)
ULONG *ptr;
{
int i;
int end;
ULONG sum;

sum=ptr[0];
end=ptr[1];
if (end>BLOCKSIZE)
  end=BLOCKSIZE;
for(i=1;i<end;i++)
  sum+=ptr[i];
return ((LONG)sum);
}

int readblock(f,nr)
int f;
int nr;
{
if (nr>=RDBS)
  return -1; 
if (lseek(f,nr*BLOCKSIZE,SEEK_SET)<0)
  return -1;
read(f,rdb+nr*BLOCKSIZE,BLOCKSIZE);
if (checksum(rdb+nr*BLOCKSIZE)!=0)
 {
  printf("block %d,cheksum %ld\n",nr,checksum(rdb+nr*BLOCKSIZE));
  return -1;
 }
if (nr>lastblock)
  lastblock=nr;
return 0;
}


int read_badblocks(f,nr)
int f,nr;
{
if (nr==-1)
  return 0;
if (readblock(f,nr)!=0)
  return -1;
if (BBB(nr)->bbb_ID!=IDNAME_BADBLOCK)
  return -1;
return read_badblocks(f,BBB(nr)->bbb_Next);
}

int read_part(f,nr)
int f,nr;
{
if (nr==-1)
  return 0;
if (readblock(f,nr)!=0)
  return -1;
if (PAB(nr)->pb_ID!=IDNAME_PARTITION)
  return -1;
return read_part(f,PAB(nr)->pb_Next);
}

int read_lseg(f,nr)
int f,nr;
{
if (nr==-1)
  return 0;
if (readblock(f,nr)!=0)
  return -1;
if (LSB(nr)->lsb_ID!=IDNAME_LOADSEG)
  return -1;
return read_lseg(f,LSB(nr)->lsb_Next);
}

int read_fshl(f,nr)
int f,nr;
{
if (nr==-1)
  return 0;
if (readblock(f,nr)!=0)
  return -1;
if (FSB(nr)->fhb_ID!=IDNAME_FILESYSHEADER)
  return -1;
if (read_lseg(f,FSB(nr)->fhb_SegListBlocks)!=0)
  return -1;
return read_fshl(f,FSB(nr)->fhb_Next);
}

int create_rdb(f)
int f;
{
struct hd_geometry geo={-1,};
char cmd[40]= {
0,0,0,0, /* input data: 0 */
0,0,0,24,/* output data: 24 */
0x1a,0,4,0,24,0, /* Mode Sense(6) page 4 (Rigid Disk) current values */
};
struct {
char PS:1; char res1:1; char PageCode:6;
char PageLength;
char NumCyls[3];
char NumHeads;
char PreComp[3];
char ReducedWrite[3];
char StepRate[2];
char LandingZone[3];
char res2:7; char RPL:1;
char Rotoffset;
char res3;
char rpm[2];
char res4[2];
} page4;
struct {
char PS:1; char res1:1; char PageCode:6;
char PageLength;
char TrackPerZone[2];
char AlSecPerZone[2];
char AlTrkPerZone[2];
char AlTrkPerLUN[2];
char SectorsPerTrack[2];
char BytesPerSector[2];
char Interleave[2];
char TrackSkew[2];
char CylSkew[2];
char SSEC:1; char HSEC:1; char RMB:1; char SURF:1; char res2:4;
char res3[3];
} page3;

int err;

if ((err=ioctl(f,SCSI_IOCTL_SEND_COMMAND,cmd))==0)
 {
  memcpy(&page4,cmd+14,sizeof(page4));
  cmd[10]=3; /* Now page 3 */
  if ((err=ioctl(f,SCSI_IOCTL_SEND_COMMAND,cmd))==0)
   {
    memcpy(&page3,cmd+14,sizeof(page3));
    if (BUILD2(page3.BytesPerSector)!=BLOCKSIZE)
     {
      printf("Blocksize differs  (%d != %d)\n",
        BUILD2(page3.BytesPerSector),BLOCKSIZE);
      return -1;
     }
   }
 }
if (err!=0)
 {
  printf("Mode Sense failed %d\n",err);
  if (ioctl(f,HDIO_GETGEO,&geo)!=0)
   {
    printf("can't get geometry data\n");
    return -1;
   }
  printf("geometry: %d heads, %d secs, %d cyl, %ld start\n",
    geo.heads,geo.sectors,geo.cylinders,geo.start);
 }
memset(rdb,-1,BLOCKSIZE);
RDB(0)->rdb_ID=IDNAME_RIGIDDISK;
RDB(0)->rdb_SummedLongs=64;
RDB(0)->rdb_BlockBytes=BLOCKSIZE;
RDB(0)->rdb_Flags=0;
RDB(0)->rdb_Cylinders=(err==0)?BUILD3(page4.NumCyls):geo.cylinders;
RDB(0)->rdb_Sectors=(err==0)?BUILD2(page3.SectorsPerTrack):geo.sectors;
RDB(0)->rdb_Heads=(err==0)?page4.NumHeads:geo.heads;
RDB(0)->rdb_Interleave=(err==0)?BUILD2(page3.Interleave):0;
RDB(0)->rdb_Park=(err==0)?BUILD3(page4.LandingZone):geo.cylinders;
RDB(0)->rdb_WritePreComp=(err==0)?BUILD3(page4.PreComp):geo.cylinders;
RDB(0)->rdb_ReducedWrite=(err==0)?BUILD3(page4.ReducedWrite):geo.cylinders;
RDB(0)->rdb_StepRate=(err==0)?BUILD2(page4.StepRate):0;
RDB(0)->rdb_RDBBlocksLo=0;
RDB(0)->rdb_RDBBlocksHi=RDB(0)->rdb_Sectors*RDB(0)->rdb_Heads*2-1;
RDB(0)->rdb_LoCylinder=2;
RDB(0)->rdb_HiCylinder=RDB(0)->rdb_Cylinders-1;
RDB(0)->rdb_CylBlocks=RDB(0)->rdb_Sectors*RDB(0)->rdb_Heads;
RDB(0)->rdb_AutoParkSeconds=0;
RDB(0)->rdb_HighRDSKBlock=0;
RDB(0)->rdb_ChkSum-=checksum(rdb);
firstblock=0;
return 0;
}

int read_rdb(dev)
char *dev;
{
int f;
int i;
int ret;

if ((f=open(dev,O_RDONLY))<0) 
 {
  printf("cant open dev\n");
  return -1;
 }
i=0;
while ((i<RDB_LOCATION_LIMIT)&&(firstblock==-1))
 {
  if ((readblock(f,i)==0)&&
      (RDB(i)->rdb_ID==IDNAME_RIGIDDISK))
    firstblock=i;
  i++;
 }
if (firstblock==-1)
 {
  printf("No valid RDB found\n");
  create_rdb(f);
  close(f);
  return 0;
 }
ret=0;
ret|=read_badblocks(f,RDB(firstblock)->rdb_BadBlockList);
ret|=read_part(f,RDB(firstblock)->rdb_PartitionList);
ret|=read_fshl(f,RDB(firstblock)->rdb_FileSysHeaderList);
ret|=read_lseg(f,RDB(firstblock)->rdb_DriveInit);
close(f);
return ret;
}

void show_rdb()
{
struct RigidDiskBlock *rb;
struct PartitionBlock *pb;
LONG block;
struct DosEnvec *de;
LONG ts;
union {
LONG dtl;
char dtc[4];
} dt;
int i;

printf("\014RDB starting at Block %d\n",firstblock);
rb=RDB(firstblock);
printf("\nRDB RDISK characteristics:\n");
printf("SummedLongs: %ld\n",rb->rdb_SummedLongs);
printf("HostID: %ld\n",rb->rdb_HostID);
printf("BlockBytes: %ld\n",rb->rdb_BlockBytes);
printf("Flags: 0x%08lx\n",rb->rdb_Flags);
printf("\nRDB hardware characteristics:\n");
printf("Cyl: %ld, Secs %ld, Heads %ld\n",
  rb->rdb_Cylinders,rb->rdb_Sectors,rb->rdb_Heads);
printf("Interleave: %ld, Park %ld\n",
  rb->rdb_Interleave,rb->rdb_Park);
printf("PreComp %ld, ReducedWrite %ld, StepRate %ld\n",
  rb->rdb_WritePreComp,rb->rdb_ReducedWrite,rb->rdb_StepRate);
printf("\nRDB logical characteristics:\n");
printf("RDBBlocksLO %ld, RDBBlocksHI %ld, LoCyl %ld, HiCyl %ld\n",
  rb->rdb_RDBBlocksLo,rb->rdb_RDBBlocksHi,rb->rdb_LoCylinder,rb->rdb_HiCylinder);
printf("CylBlocks %ld, AutoPark in %lds, HighRDSKBlock %ld\n",
  rb->rdb_CylBlocks,rb->rdb_AutoParkSeconds,rb->rdb_HighRDSKBlock);
printf("\nPartion data:\n");
block=rb->rdb_PartitionList;
while(block!=-1) 
 {
  printf("partion at block %ld\n",block);
  pb=PAB(block);
  printf("Flags 0x%08lx (%sboot, %smount)\n",pb->pb_Flags,
    (pb->pb_Flags&PBFF_BOOTABLE)?"":"no",
    (pb->pb_Flags&PBFF_NOMOUNT)?"no":"");
  printf("DevFlags 0x%08lx\n",pb->pb_DevFlags);
  printf("DriveName: ");
  for(i=1;i<=pb->pb_DriveName[0];i++)
    printf("%c",pb->pb_DriveName[i]);
  printf("\n");
  de=(struct DosEnvec *)pb->pb_Environment;
  printf("Enviroment Size: %ld\n",de->de_TableSize);
  ts=de->de_TableSize;
  printf("Surfaces : %ld\n",de->de_Surfaces);
  printf("Blocks/Track: %ld\n",de->de_BlocksPerTrack);
  printf("LowCyl: %ld, HighCyl: %ld\n",de->de_LowCyl,de->de_HighCyl);
  printf("Size in KBytes: %ld\n",
    (de->de_HighCyl-de->de_LowCyl+1)*de->de_BlocksPerTrack*de->de_Surfaces/2);
  if (ts<12)
    de->de_BufMemType=0;
  if (ts<13)
    de->de_MaxTransfer=0x7fffffff;
  if (ts<14)
    de->de_Mask=0xffffffff;
  if (ts<15)
    de->de_BootPri=0;
  printf("BootPri: %ld\n",de->de_BootPri);
  if (ts<16)
    de->de_DosType=0x444f5300;
  dt.dtl=de->de_DosType;
  printf("DosType: 0x%08lx (",dt.dtl);
  for(i=0;i<4;i++)
    printf(isprint(dt.dtc[i])?"%c":"\\%03o",dt.dtc[i]);
  printf(")\n");
  if (ts<17)
    de->de_Baud=0;
  if (ts<18)
    de->de_Control=0;
  if (ts<19)
   {
    de->de_BootBlocks=0;
    de->de_TableSize=19;
   }
  printf("BootBlocks: %ld\n",de->de_BootBlocks);
  block=pb->pb_Next;
 }
}

int cp_rdbs(to, block, from, offset)
u_char *to;
int block,from, offset;
{
int cb;

if (from==-1)
  return 0;
cb=0;
memcpy(to,LSB(from),BLOCKSIZE);
if (LSB(from)->lsb_Next!=-1)
 {
  cb=cp_rdbs(to+BLOCKSIZE,block+1,LSB(from)->lsb_Next,offset);
  ((struct LoadSegBlock *)to)->lsb_Next=block+1+offset;
 }
((struct LoadSegBlock *)to)->lsb_ChkSum-=checksum(to);
return cb+1;
}

int cp_fshd(to, block, from, offset)
struct FileSysHeaderBlock *to;
int block, from, offset;
{
int cb;

if (from==-1)
  return 0;
memcpy(to,FSB(from),BLOCKSIZE);
cb=cp_rdbs(((u_char *)to)+BLOCKSIZE,block+1,to->fhb_SegListBlocks,offset);
to->fhb_SegListBlocks=(cb==0)?-1:block+1+offset;
if (FSB(from)->fhb_Next!=-1)
 {
  cb+=cp_fshd(((u_char *)to)+BLOCKSIZE*(cb+1),block+cb+1,FSB(from)->fhb_Next,offset);
  to->fhb_Next=block+cb+1+offset;
 }
to->fhb_ChkSum-=checksum(to);
return cb+1;
}

int save_rdb(dev,block)
char *dev;
int block;
{
int f;
u_char *buf;
int blocks,cb;

if (block>RDB_LOCATION_LIMIT)
  return -1;
if ((buf=malloc((RDB(firstblock)->rdb_HighRDSKBlock-firstblock+1)*BLOCKSIZE))==NULL)
  return -1;
memcpy(buf,rdb+firstblock*BLOCKSIZE,BLOCKSIZE);
blocks=1;
cb=cp_rdbs(buf+BLOCKSIZE,1,RDB(firstblock)->rdb_BadBlockList,block); 
((struct RigidDiskBlock *)buf)->rdb_BadBlockList=(cb==0)?-1:blocks+block;
blocks+=cb;
cb=cp_rdbs(buf+BLOCKSIZE*blocks,blocks,RDB(firstblock)->rdb_PartitionList,block);
((struct RigidDiskBlock *)buf)->rdb_PartitionList=(cb==0)?-1:blocks+block;
blocks+=cb;
cb=cp_fshd(buf+BLOCKSIZE*blocks,blocks,RDB(firstblock)->rdb_FileSysHeaderList,block);
((struct RigidDiskBlock *)buf)->rdb_FileSysHeaderList=(cb==0)?-1:blocks+block;
blocks+=cb;
cb=cp_rdbs(buf+BLOCKSIZE*blocks,blocks,RDB(firstblock)->rdb_DriveInit,block);
((struct RigidDiskBlock *)buf)->rdb_DriveInit=(cb==0)?-1:blocks+block;
blocks+=cb;
((struct RigidDiskBlock *)buf)->rdb_HighRDSKBlock=blocks+block;
((struct RigidDiskBlock *)buf)->rdb_ChkSum-=checksum(buf);
if ((f=open(dev,O_WRONLY))<0)
 {
  free(buf);
  return -1;
 }
lseek(f,block*BLOCKSIZE,SEEK_SET);
write(f,buf,blocks*BLOCKSIZE);
close(f);
free(buf);
return 0;
}

int main(argc,argv)
int argc;
char **argv;
{
char *dev="/dev/hda";
char a[2];
int done;
struct PartitionBlock *pa[PARTS];
int i,j,parts;
struct DosEnvec *de;

if (argc>1)
  dev=argv[1];

if ((rdb=malloc(BLOCKSIZE*RDBS))==NULL)
 {
  printf("no mem\n");
  exit(10);
 }

if(read_rdb(dev)!=0)
 {
  printf("read error\n");
  exit(20);
 }
show_rdb();
if (RDB(firstblock)->rdb_HighRDSKBlock==-1)
  RDB(firstblock)->rdb_HighRDSKBlock=lastblock;
i=RDB(firstblock)->rdb_PartitionList;
parts=0;
while(i!=-1)
 {
  pa[parts]=PAB(i);
  i=PAB(i)->pb_Next;
  parts++;
  if (parts>=PARTS)
   {
    printf("too many parts. Please recompile\n");
    exit(10);
   }
 }
done=0;
while(done==0)
 {
  printf("\n\n? :");
  scanf("%c%*c",a,a);
  switch(a[0])
   {
    case 'q':
    case 'x':
      done=1;
      break;
    case 's':
      printf("Nr. LoCyl HiCyl Pri b,m  DosType   CBB\n");
      printf("--------------------------------------\n");
      for(i=0;i<parts;i++)
       {
        de=(struct DosEnvec *)pa[i]->pb_Environment;
        printf("%2d: %5ld %5ld %3ld %c,%c 0x%08lx %3ld\n",
          i+1,de->de_LowCyl,de->de_HighCyl,de->de_BootPri,
          pa[i]->pb_Flags&PBFF_BOOTABLE?'b':'x',
          pa[i]->pb_Flags&PBFF_NOMOUNT?'x':'m',
          de->de_DosType,de->de_BootBlocks);
       }
      break;
    case 'p':
      i=0;
      while ((i<1)||(i>parts))
       {
	printf("Nr.: ");
	scanf("%d%*c",&i,a);
       }
      i--;
      de=(struct DosEnvec *)pa[i]->pb_Environment;
      printf("?? :");
      scanf("%c%*c",a,a);
      switch(a[0])
       {
        case 'l':
         printf("LoCyl: ");
         scanf("%ld",&de->de_LowCyl);
         break;
        case 'h':
          printf("HiCyl: ");
          scanf("%ld",&de->de_HighCyl);
          break;
        case 'p':
          printf("Pri: ");
          scanf("%ld",&de->de_BootPri);
          break;
        case 'b':
          pa[i]->pb_Flags^=PBFF_BOOTABLE;
          break;
        case 'm':
          pa[i]->pb_Flags^=PBFF_NOMOUNT;
          break;
        case 'd':
          printf("DosType: ");
          scanf("%li",&de->de_DosType);
          break;
        case 'c':
          printf("BootBlocks: ");
          scanf("%ld",&de->de_BootBlocks);
          break;
        default:
          printf("def %c\n",a[0]);
          break;
       }
      break;
    case 'd':
      if (parts>0)
       {
	i=0;
	while((i<1)||(i>parts))
	 {
	  printf("Nr.: ");
	  scanf("%d%*c",&i,a);
	 }
	i--;
	if (i==0)
	  RDB(firstblock)->rdb_PartitionList=pa[0]->pb_Next;
	else
	  pa[i-1]->pb_Next=pa[i]->pb_Next;
	while(i<parts)
	 {
	  pa[i]=pa[i+1];
	  i++;
	 }
	parts--;
       }
      break;
    case 'n':
      i=0;
      if (parts>0)
       {
	while((i<1)||(i>parts+1))
	 {
	  printf("Nr.: ");
	  scanf("%d%*c",&i,a);
	 }
	i--;
       }
      if (RDB(firstblock)->rdb_HighRDSKBlock==RDB(firstblock)->rdb_RDBBlocksHi)
       {
        printf("no space\n");
        exit(10);
       }
      if (parts==PARTS-1)
       {
        printf("too many parts. please recompile\n");
        exit(10);
       }
      for(j=parts;j>i;j--)
	pa[j]=pa[j-1];
      parts++;
      pa[i]=PAB(++RDB(firstblock)->rdb_HighRDSKBlock);
      memset(pa[i],-1,BLOCKSIZE);
      pa[i]->pb_ID=IDNAME_PARTITION;
      pa[i]->pb_SummedLongs=64;
      pa[i]->pb_HostID=RDB(firstblock)->rdb_HostID;
      if (i==0)
       {
        pa[i]->pb_Next=RDB(firstblock)->rdb_PartitionList;
        RDB(firstblock)->rdb_PartitionList=RDB(firstblock)->rdb_HighRDSKBlock;
       }
      else
       {
        pa[i]->pb_Next=pa[i-1]->pb_Next;
        pa[i-1]->pb_Next=RDB(firstblock)->rdb_HighRDSKBlock;
       }
      pa[i]->pb_Flags=0;
      pa[i]->pb_DevFlags=0;
      memcpy(pa[i]->pb_DriveName,"\013Amiga_Fdisk",12);;
      de=(struct DosEnvec *)pa[i]->pb_Environment;
      de->de_TableSize=19;
      de->de_SizeBlock=128;
      de->de_SecOrg=0;
      de->de_Surfaces=RDB(firstblock)->rdb_Heads;
      de->de_SectorPerBlock=1;
      de->de_BlocksPerTrack=RDB(firstblock)->rdb_Sectors;
      de->de_Reserved=2;
      de->de_PreAlloc=0;
      de->de_Interleave=0;
      de->de_NumBuffers=30;
      de->de_BufMemType=0;
      de->de_MaxTransfer=0x7fffffff;
      de->de_Mask=0xffffffff;
      de->de_Baud=0;
      de->de_Control=0;
      break;
    case 'w':
      printf("First Block: ");
      scanf("%d%*c",&i,a);
      if (save_rdb(dev,i)!=0)
        printf("Saving failed. RDB may be unusuable.\n");
      break;
   }
 }
free(rdb);
return 0;
}

